"""
Module encapsulating message structure to facilitate use with the Consumer and Producer wrappers
"""
from dataclasses import asdict
import json
from typing import Union, Iterable

import pydantic

__all__ = ["message_class", "MessageValidationException"]


class MessageValidationException(Exception):
    """
    Exception raised when a message cannot be parsed
    """


def message_class(routing_key: str, queues: Union[Iterable[str], None] = None, **kwargs):
    """
    Standard lib dataclasses.dataclass like decorator for creating message classes
    :param routing_key: Routing key of the message e.g. message.m
    :param queues: Iterable of queue names to bind the routing key to e.g. [message.q,message2.q]
    :param kwargs: keyword arguments passed through to the pydantic.dataclasses.dataclass function
    :return: A pydantic.dataclasses.dataclass for the Message with the following methods
        - routing_key() : returns the str routing key
        - binding(): returns a list of dicts specifying the routing key to queue binding
        - from_dict(body: dict): Convert a dict message body to Message instance
        - from_bytes(body: bytes): Convert a bytes message body to Message instance
        - dict(**kwargs): Convert a dict message body to Message instance
    """
    def to_dict(self, **kwargs):
        """
        convert dataclass instance to a dict
        :param kwargs: keyword arguments passed through to the dataclasses.asdict function
        :return: dict of the instance
        """
        return asdict(self, **kwargs)

    @classmethod
    def from_dict(cls, body: dict):
        """
        Convert a dict message body to Message instance
        :param body: The string body of the message to parse into a Message instance
        :return: instance of the message class
        :raises: MessageValidationException, TypeError
        """
        if not isinstance(body, dict):
            raise TypeError("Message must be a dict to validate schema")
        valid_args = {k: v for k, v in body.items() if k in cls.__dataclass_fields__.keys()}
        try:
            return cls(**valid_args)
        except (UnicodeDecodeError, TypeError, ValueError) as e:  # Catch parsing Failure
            raise MessageValidationException("Message schema validation failed") from e

    @classmethod
    def from_bytes(cls, body: bytes):
        """
        Convert a bytes message body to Message instance
        :param body: The string body of the message to parse into a Message instance
        :return: instance of the message class
        :raises: MessageValidationException
        """
        try:
            # convert to dict
            body = json.loads(body)
            # validate against schema definition
            message = cls.from_dict(body)
        except (UnicodeDecodeError, TypeError, ValueError) as e:  # Catch parsing Failure
            raise MessageValidationException("Message schema validation failed") from e
        return message

    if queues is None:
        queues = list()

    def decorator(cls):
        """
        Decorator implementation for the message class
        """

        @classmethod
        def routing_key_method(cls):
            return routing_key

        @classmethod
        def binding_method(cls):
            return [{'routing_key': cls.routing_key(), 'bound_queue': q} for q in queues]
        MessageClass = pydantic.dataclasses.dataclass(cls, **kwargs)
        MessageClass.routing_key = routing_key_method
        MessageClass.binding = binding_method
        MessageClass.dict = to_dict
        MessageClass.from_dict = from_dict
        MessageClass.from_bytes = from_bytes

        return MessageClass
    return decorator



"""
Tests the producer wrapper
"""
import json
import pytest

from talus import (DurableBlockingProducerWrapper,
                   DurableBlockingConsumerWrapper)


ROUTING_KEY = 'test.message.m'
QUEUE_NAME = 'test.queue.q'
BINDINGS = [{'routing_key': ROUTING_KEY, 'bound_queue': QUEUE_NAME},]
EXCHANGE = 'amq.direct'


@pytest.fixture
def producer():
    """
    Fixture producing an instance of the DurableBlockingProducerWrapper.  Disconnects on exit
    :return: DurableBlockingProducerWrapper
    """
    producer_wrapper = DurableBlockingProducerWrapper(BINDINGS, EXCHANGE)
    yield producer_wrapper
    producer_wrapper.disconnect()


@pytest.fixture
def consumer():
    """
    Fixture producing an instance of the DurableBlockingConsumerWrapper.  Disconnects on exit
    :return: DurableBlockingConsumerWrapper
    """
    consumer_wrapper = DurableBlockingConsumerWrapper(QUEUE_NAME)
    yield consumer_wrapper
    consumer_wrapper.disconnect()


@pytest.fixture
def empty_queue(consumer):
    """
    Fixture that empties the test queue
    :param consumer: Fixture returning Instance of the DurableBlockingConsumerWrapper
    :return: None
    """
    for method, properties, body in consumer.consume_generator(auto_ack=True):
        if method is None:
            consumer.cancel_consume_generator()
    return


@pytest.fixture
def single_message_on_queue(producer, empty_queue):
    """
    Fixture that produces a single message to an empty test queue
    :param producer: Fixture returning Instance of the DurableBlockingProducerWrapper
    :return: dict of the message that was placed on the queue
    """
    message = {'test_case': 'message_to_consume fixture'}
    producer.post(ROUTING_KEY, message)
    return message


def test_post_message(producer, consumer, empty_queue):
    """
    Given - An instantiated producer and consumer and an empty test queue
    When - Post a message to the test queue
    Then - Message consumed from the test queue matches the posted one.
    :param producer: Fixture returning Instance of the DurableBlockingProducerWrapper
    :param consumer: Fixture returning Instance of the DurableBlockingConsumerWrapper
    :param empty_queue: Fixture returning None that ensured an empty test queue
    :return: None
    """
    message = {"test_case": "test_post_message"}
    producer.post(ROUTING_KEY, message)

    for method, properties, body in consumer.consume_generator(auto_ack=True):
        assert (json.loads(body) == message)
        break


def test_consume_acknowledge(consumer, single_message_on_queue):
    """
    Given - An consumer instance and a message on a queue
    When - Consumer consumes and acknowledges message
    Then - Queue is empty
    :param consumer: Fixture returning Instance of the DurableBlockingConsumerWrapper
    :param single_message_on_queue: Fixture returning a dict body of a single message on the test queue
    :return: None
    """
    for i, (method, properties, body) in enumerate(consumer.consume_generator(auto_ack=False)):
        if i == 0:
            assert (json.loads(body) == single_message_on_queue)
            consumer.acknowledge_message(method.delivery_tag)
        if i == 1:
            assert body is None
            break


def test_consume_reject(consumer, single_message_on_queue):
    """
    Given - An consumer instance and a message on a queue
    When - Consumer consumes and rejects message
    Then - Queue is empty
    :param consumer: Fixture returning Instance of the DurableBlockingConsumerWrapper
    :param single_message_on_queue: Fixture returning a dict body of a single message on the test queue
    :return: None
    """
    for i, (method, properties, body) in enumerate(consumer.consume_generator(auto_ack=False)):
        if i == 0:
            assert (json.loads(body) == single_message_on_queue)
            consumer.reject_message(method.delivery_tag)
        if i == 1:
            assert body is None
            break


def test_consume_requeue(consumer, single_message_on_queue):
    """
    Given - An consumer instance and a message on a queue
    When - Consumer consumes and requeues message
    Then - Queue has the requeued message
    :param consumer: Fixture returning Instance of the DurableBlockingConsumerWrapper
    :param single_message_on_queue: Fixture returning a dict body of a single message on the test queue
    :return: None
    """
    for i, (method, properties, body) in enumerate(consumer.consume_generator(auto_ack=False)):
        if i == 0:
            assert (json.loads(body) == single_message_on_queue)
            consumer.requeue_message(method.delivery_tag)
        if i == 1:
            assert (json.loads(body) == single_message_on_queue)
            break

def test_listen(consumer, single_message_on_queue):
    """
    Given - A consumer instance and a single message on the queue
    When - Listening for messages on the queue
    Then - Single message is consumed
    :param consumer: Fixture returning Instance of the DurableBlockingConsumerWrapper
    :param single_message_on_queue: Fixture returning a dict body of a single message on the test queue
    :return: None
    """
    def verify_message(ch, method, properties, body):
        assert (json.loads(body) == single_message_on_queue)
        ch.basic_ack(method.delivery_tag)
        consumer.disconnect()

    consumer.listen(verify_message)


def test_context(single_message_on_queue):
    """
    Given - A single message on a queue
    When - When instantiating the consumer using 'with' and consuming a message
    Then - Consumer is connected inside the with and disconnected outside
    :return: None
    """
    with DurableBlockingConsumerWrapper(QUEUE_NAME) as consumer:
        for method, properties, body in consumer.consume_generator(auto_ack=True):
            assert (json.loads(body) == single_message_on_queue)
            break
        assert consumer.is_connected

    assert not consumer.is_connected


def test_invalid_routing_key(producer):
    """
    Given - A producer instance
    When - Post a message with a non-configured routing key
    Then - raise a value error
    :param producer: Fixture returning Instance of the DurableBlockingProducerWrapper
    :return: None
    """
    with pytest.raises(ValueError):
        producer.post('bad_key', {'foo': 'bar'})


def test_listen_connect(producer, consumer):
    """
    Given - A producer and consumer instance
    When - A message is posted and consumed first by listen
    Then - the consumer is connected.
    :param producer: Fixture returning Instance of the DurableBlockingProducerWrapper
    :param consumer: Fixture returning Instance of the DurableBlockingConsumerWrapper
    :return: None
    """
    producer.post(ROUTING_KEY, {'foo': 'bar'})

    def callback(ch, method, properties, body):
        ch.basic_ack(method.delivery_tag)
        assert consumer.is_connected
        consumer.disconnect()

    consumer.listen(callback)



"""
Unit tests for message parsing
"""
import json

import pytest

from talus.message import (
    message_class,
    MessageValidationException,
)


@message_class(routing_key="example.m", queues=["example1.q", "example2.q"])
class OutboundMessage:
    s: str
    i: int
    f: float
    d: str = "default"


@message_class(routing_key="example.m")
class InboundMessage:
    s: str
    i: int
    f: float
    d: str = "default"


@pytest.fixture(params=["inbound", "outbound"])
def MessageClass(request):
    message_classes = {
        "inbound": InboundMessage,
        "outbound": OutboundMessage
    }
    return message_classes[request.param]


@pytest.fixture(params=['all', 'all_required', 'coerce_int', 'coerce_float', 'coerce_str'])
def valid_message_body(request) -> dict:
    message_bodies = {
        "all": {"s": "s", "i": 1, "f": 1.1, "d": "d"},
        "all_required": {"s": "s", "i": 1, "f": 1.1},
        "coerce_int": {"s": "s", "i": "1", "f": 1.1},
        "coerce_float": {"s": "s", "i": 1, "f": "1.1"},
        "coerce_str": {"s": 1, "i": 1, "f": 1.1},
    }
    return message_bodies[request.param]


def test_valid_message_from_bytes(valid_message_body, MessageClass):
    """
    Given: valid message body
    When: parsing the message
    Then: Get a a valid message instance
    """
    body = json.dumps(valid_message_body).encode('utf-8')
    message = MessageClass.from_bytes(body)
    assert message.s == str(valid_message_body['s'])
    assert message.i == int(valid_message_body['i'])
    assert message.f == float(valid_message_body['f'])
    assert message.d == str(valid_message_body.get('d', 'default'))
    assert message.routing_key() == "example.m"
    if isinstance(message, OutboundMessage):
        assert message.binding() == [
            {
                "routing_key": "example.m",
                "bound_queue": "example1.q"
            },
            {
                "routing_key": "example.m",
                "bound_queue": "example2.q"
            },
        ]


@pytest.fixture(params=["bytes", "string", "missing_required"])
def invalid_message_body(request) -> dict:
    message_bodies = {
        "bytes": b"bytes string",
        "string": "just a string",
        "missing_required": json.dumps({"i": "1", "f": 1.1}).encode('utf-8'),
    }
    return message_bodies[request.param]


def test_invalid_message_from_bytes(invalid_message_body, MessageClass):
    """
    Given: invalid message body
    When: body is parsed
    Then: MessageValidationException is raised
    """
    with pytest.raises(MessageValidationException):
        MessageClass.from_bytes(invalid_message_body)


def test_message_dict(MessageClass):
    """
    Given: valid dict message body
    When: message instance is created dict
    Then: message instance dict method returns the input dict
    """
    body = {"s": "s", "i": 1, "f": 1.1, "d": "d"}
    message = MessageClass.from_dict(body)
    assert message.dict() == body



"""
Base wrapper definition shared by both consumer and producer wrappers
"""
import logging

import pika
from pika.exceptions import (
    ConnectionOpenAborted,
    StreamLostError,
    NoFreeChannels,
    ConnectionWrongStateError,
    ConnectionClosed,
    ConnectionClosedByClient,
    ConnectionBlockedTimeout,
    AMQPHeartbeatTimeout,
    ChannelWrongStateError,
    ChannelClosed,
    ChannelClosedByClient,
    InvalidChannelNumber,
    UnexpectedFrameError,
    ChannelError
)

from retry.api import retry_call


logger = logging.getLogger(__name__)
SECONDS_60 = 60


class DurableBlockingBaseWrapper:
    """
    RabbitMQ connector that establishes a blocking connection and channel
    """
    # exceptions that will cause a retry including the connect cycle
    retry_exceptions = (
        ConnectionOpenAborted,
        StreamLostError,
        NoFreeChannels,
        ConnectionWrongStateError,
        ConnectionClosedByClient,
        ConnectionBlockedTimeout,
        AMQPHeartbeatTimeout,
        ChannelWrongStateError,
        ChannelClosed,
        ChannelClosedByClient,
        InvalidChannelNumber,
        UnexpectedFrameError,
        ChannelError
    )

    def __init__(
        self,
        rabbitmq_host: str,
        rabbitmq_port: int,
        rabbitmq_user: str,
        rabbitmq_pass: str,
        retry_delay: int,
        retry_backoff: int,
        retry_jitter: int,
        retry_max_delay: int,
        retry_tries: int,
        connection_name: str,
        prefetch_count: int,
        heartbeat: int = SECONDS_60
    ):
        """
        Constructor for the base connector

        :param rabbitmq_host: Host name or IP of the rabbitMQ server. e.g. 127.0.0.1

        :param rabbitmq_port: Port the rabbitmq server listens on e.g. 5672

        :param rabbitmq_user: Username for the rabbitMQ server e.g. guest

        :param rabbitmq_pass: Password for the rabbitMQ server e.g. guest

        :param retry_delay: initial delay between attempts.

        :param retry_backoff: multiplier applied to delay between attempts.

        :param retry_jitter: extra seconds added to delay between attempts.
                   fixed if a number, random if a range tuple (min, max)

        :param retry_max_delay: the maximum value of delay.

        :param connection_name: Name of the connection that will be visible in the rabbitmq admin console

        :param prefetch_count: Number of un-acked messages that can be consumed
        """
        self.connection_parameters = pika.ConnectionParameters(
            host=rabbitmq_host,
            port=rabbitmq_port,
            credentials=pika.credentials.PlainCredentials(rabbitmq_user, rabbitmq_pass),
            heartbeat=heartbeat,
            client_properties={"connection_name": connection_name},
        )
        self.connection = None
        self.channel = None
        self.prefetch_count = prefetch_count
        self.retry_config = {
            "exceptions": self.retry_exceptions,
            "delay": retry_delay,
            "backoff": retry_backoff,
            "jitter": retry_jitter,
            "max_delay": retry_max_delay,
            "tries": retry_tries,
        }

    @property
    def is_connected(self):
        """
        Current state of the connection.  Only updated when the connection is used.

        :return: Latest connection state
        """
        if self.connection:
            return self.connection.is_open
        return False

    def _connect(self):
        """
        Configures and initiates connection to the RabbitMQ server.

        :return: None
        """
        logger.debug(
            f"Attempt to connect to RabbitMQ: connection_params={self.connection_parameters}"
        )
        self.connection = pika.BlockingConnection(self.connection_parameters)
        logger.info(f"Connection Created")
        self.channel = self.connection.channel()
        logger.info("Channel Created")
        self.channel.confirm_delivery()  # ensure persistence prior to message confirmation
        self.channel.basic_qos(
            prefetch_count=self.prefetch_count
        )  # Number of un-Acked message delivered at a time
        logger.info("Channel configured")

    def connect(self):
        """
        Retries as configured the connection to the RabbitMQ server.

        :return: None
        """
        retry_call(self._connect, **self.retry_config)

    def disconnect(self):
        """
        Closes connection and related channels to the RabbitMQ Server.

        :return: None
        """
        if self.is_connected:
            self.connection.close()
        logger.warning(f"Disconnected from RabbitMQ: " f"connection={self.connection_parameters}")

    def __enter__(self):
        """
        Entry for with context manager.

        :return: connected instance of self
        """
        self.connect()
        return self

    def __exit__(self, exc_type, value, traceback):
        """
        Exit for with context manager.  Disconnects from rabbitmq

        :return: None
        """
        self.disconnect()



"""
Consumer implementation of the base connection wrapper
"""

import logging
from typing import Callable, Generator

from retry.api import retry_call

from talus.base import DurableBlockingBaseWrapper


logger = logging.getLogger(__name__)


__all__ = ["DurableBlockingConsumerWrapper"]


class DurableBlockingConsumerWrapper(DurableBlockingBaseWrapper):
    """
    RabbitMQ connector for consuming from a single queue in RabbitMQ.
    """

    def __init__(
        self,
        consumer_queue: str,
        rabbitmq_host: str = "localhost",
        rabbitmq_port: int = 5672,
        rabbitmq_user: str = "guest",
        rabbitmq_pass: str = "guest",
        retry_delay: int = 1,
        retry_backoff: int = 2,
        retry_jitter: int = (1, 10),
        retry_max_delay: int = 300,
        retry_tries: int = -1,
        prefetch_count: int = 1,
        connection_name: str = "consumer connection",
    ):
        """
        Constructor for the consumer connector

        :param consumer_queue: Name of the queue the consumer will listen for messages on.

        :param rabbitmq_host: Host name or IP of the rabbitMQ server. e.g. 127.0.0.1

        :param rabbitmq_port: Port the rabbitmq server listens on e.g. 5674

        :param rabbitmq_user: Username for the rabbitMQ server e.g. guest

        :param rabbitmq_pass: Password for the rabbitMQ server e.g. guest

        :param retry_delay: initial delay between attempts.

        :param retry_backoff: multiplier applied to delay between attempts.

        :param retry_jitter: extra seconds added to delay between attempts.
                   fixed if a number, random if a range tuple (min, max)

        :param retry_max_delay: the maximum value of delay.

        :param connection_name: Name of the connection that will be visible in the rabbitmq admin console

        :param prefetch_count: Number of un-acked messages that can be consumed
        """
        super().__init__(
            rabbitmq_host=rabbitmq_host,
            rabbitmq_port=rabbitmq_port,
            rabbitmq_user=rabbitmq_user,
            rabbitmq_pass=rabbitmq_pass,
            retry_delay=retry_delay,
            retry_backoff=retry_backoff,
            retry_jitter=retry_jitter,
            retry_max_delay=retry_max_delay,
            retry_tries=retry_tries,
            prefetch_count=prefetch_count,
            connection_name=connection_name,
        )
        self.consumer_queue = consumer_queue

    def _connect_consumer(self) -> None:
        """
        Configures and initiates consumer connection to the RabbitMQ server.
        :return:
        """
        if not self.is_connected:
            super().connect()

        self.channel.queue_declare(queue=self.consumer_queue, durable=True)
        logger.info(f"Queue Created: queue={self.consumer_queue}")

    def connect(self) -> None:
        """
        Retries configuration of consumer connection to the RabbitMQ server
        :return:
        """
        retry_call(self._connect_consumer, **self.retry_config)
        logger.info(f"Connected to RabbitMQ: " f"connection={self.connection_parameters}")

    def _listen(self, callback: Callable) -> None:
        """
        Listens for messages on the channel configured on the consumer instance

        :param callback: Function to execute when a message is recieved. with the signature
        (ch, method, properties, body).
        ch: Copy of the channel used to acknowledge receipt (pika.Channel)
        method: Management keys for the delivered message e.g. delivery mode (pika.spec.Basic.Deliver)
        properties: Message properties (pika.spec.BasicProperties)
        body: Message body for a transfer message (bytes)

        :return: None
        """
        if not self.is_connected:
            self.connect()
        logger.info(f"Starting Listener on Queue: consumer_queue={self.consumer_queue}")
        self.channel.basic_consume(queue=self.consumer_queue, on_message_callback=callback)
        self.channel.start_consuming()

    def listen(self, callback: Callable) -> None:
        """
        Retries calls to _listen

        :param callback: Function to execute when a message is recieved. with the signature
        (ch, method, properties, body).
        ch: Copy of the channel used to acknowledge receipt (pika.Channel)
        method: Management keys for the delivered message e.g. delivery mode (pika.spec.Basic.Deliver)
        properties: Message properties (pika.spec.BasicProperties)
        body: Message body for a transfer message (bytes)

        :return: None
        """
        retry_call(self._listen, fargs=[callback], **self.retry_config)

    def consume_generator(self, auto_ack=False, inactivity_timeout: float = .1) -> Generator:
        """
        Creates a generator for messages that are on the instance consumer_queue.
        Retry logic is not applied to prevent the resetting of the generator cursor

        :param inactivity_timeout:

        :return: Generator of (method, properties, body)
        """
        if not self.is_connected:
            self.connect()
        logger.info(f"Creating consumer generator on Queue: consumer_queue={self.consumer_queue}")
        return self.channel.consume(
            queue=self.consumer_queue, auto_ack=auto_ack, inactivity_timeout=inactivity_timeout
        )

    def cancel_consume_generator(self) -> None:
        """
        Resets the active consume generator
        :return: None
        """
        logger.info(f"Cancelling consumer generator on Queue: consumer_queue={self.consumer_queue}")
        self.channel.cancel()

    def acknowledge_message(self, delivery_tag, multiple=False) -> None:
        """
        Record a message as acknowledged.
        Retry logic is not applied since creating a new channel would be unable
        to acknowledge the message recieved on the now dead channel

        :param delivery_tag: method.delivery_tag

        :param multiple:

        :return: None
        """
        self.channel.basic_ack(delivery_tag, multiple)

    def reject_message(self, delivery_tag) -> None:
        """
        Record a message as rejected.  Will go to dead letter exchange if configured on the server.
        Retry logic is not applied since creating a new channel would be unable
        to acknowledge the message recieved on the now dead channel

        :param delivery_tag: method.delivery_tag

        :return: None
        """
        self.channel.basic_reject(delivery_tag=delivery_tag, requeue=False)

    def requeue_message(self, delivery_tag) -> None:
        """
        Return message back to the queue.
        Retry logic is not applied since creating a new channel would be unable
        to acknowledge the message recieved on the now dead channel

        :param delivery_tag: method.delivery_tag

        :return: None
        """
        self.channel.basic_nack(delivery_tag=delivery_tag, requeue=True)



"""
Producer implementation of the base connection wrapper
"""

import json
import logging
from typing import List, Union

import pika
from pika.exceptions import UnroutableError, NackError
from retry import retry
from retry.api import retry_call

from talus.base import DurableBlockingBaseWrapper

logger = logging.getLogger(__name__)


__all__ = ["DurableBlockingProducerWrapper"]

HOUR_1 = 3600

class DurableBlockingProducerWrapper(DurableBlockingBaseWrapper):
    """
    RabbitMQ Connector for posting to 1 to many queues via an exchange
    """

    def __init__(
        self,
        producer_queue_bindings: Union[List[dict], dict],
        publish_exchange: str,
        rabbitmq_host: str = "localhost",
        rabbitmq_port: int = 5672,
        rabbitmq_user: str = "guest",
        rabbitmq_pass: str = "guest",
        retry_delay: int = 1,
        retry_backoff: int = 2,
        retry_jitter: int = (1, 10),
        retry_max_delay: int = 300,
        retry_tries: int = -1,
        prefetch_count: int = 1,
        connection_name: str = "producer connection",
    ):
        """
        Constructor for the producer connector

        :param producer_queue_bindings: Bindings from routing key to destination queue for an exchange
        e.g. [{"routing_key": "frame.audit.m", "bound_queue": "data.holding.audit.q"},]

        :param publish_exchange: Name of the exchange that the  producer will publish to.

        :param rabbitmq_host: Host name or IP of the rabbitMQ server. e.g. 127.0.0.1

        :param rabbitmq_port: Port the rabbitmq server listens on e.g. 5674

        :param rabbitmq_user: Username for the rabbitMQ server e.g. guest

        :param rabbitmq_pass: Password for the rabbitMQ server e.g. guest

        :param retry_delay: initial delay between attempts.

        :param retry_backoff: multiplier applied to delay between attempts.

        :param retry_jitter: extra seconds added to delay between attempts.
                   fixed if a number, random if a range tuple (min, max)

        :param retry_max_delay: the maximum value of delay.

        :param prefetch_count: Number of un-acked messages that can be consumed

        :param connection_name: Name of the connection that will be visible in the rabbitmq admin console
        """
        super().__init__(
            rabbitmq_host=rabbitmq_host,
            rabbitmq_port=rabbitmq_port,
            rabbitmq_user=rabbitmq_user,
            rabbitmq_pass=rabbitmq_pass,
            retry_delay=retry_delay,
            retry_backoff=retry_backoff,
            retry_jitter=retry_jitter,
            retry_max_delay=retry_max_delay,
            retry_tries=retry_tries,
            connection_name=connection_name,
            prefetch_count=prefetch_count,
            heartbeat=HOUR_1
        )
        if isinstance(producer_queue_bindings, dict):
            producer_queue_bindings = [producer_queue_bindings]
        self.producer_queue_bindings = producer_queue_bindings
        self.publish_exchange = publish_exchange
        self.publish_message_properties = pika.BasicProperties(
            content_type="text/plain", priority=0, delivery_mode=2, content_encoding="UTF-8"
        )

    def _connect_producer(self) -> None:
        """
        Configures and initiates producer connection to the RabbitMQ server.
        :return:
        """
        if not self.is_connected:
            super().connect()

        for queue_binding in self.producer_queue_bindings:
            self.channel.queue_declare(queue=queue_binding["bound_queue"], durable=True)
            logger.info(f"Queue Created: queue={queue_binding['bound_queue']}")
            self.channel.queue_bind(
                exchange=self.publish_exchange,
                queue=queue_binding["bound_queue"],
                routing_key=queue_binding["routing_key"],
            )
            logger.info(
                f"Bindings configured: exchange={self.publish_exchange}, "
                f"queue={queue_binding['bound_queue']}, "
                f"routing_key={queue_binding['routing_key']} "
            )

    def connect(self) -> None:
        """
        Retries configuration of producer connection to the RabbitMQ server
        :return:None
        """
        retry_call(self._connect_producer, **self.retry_config)
        logger.info(f"Connected to RabbitMQ: " f"connection={self.connection_parameters}")

    def _validate_routing_key(self, routing_key: str):
        """
        Validate that the routing key is configured on the instance to prevent un-route-able posts
        :param routing_key: key to check for the existence of
        :return: True if the routing key exists on the instance. Raise a value error if it does not
        """
        if any([routing_key in binding["routing_key"] for binding in self.producer_queue_bindings]):
            return True

        raise ValueError(
            f"Routing key not configured: "
            f"routing_key={routing_key}, "
            f"initialized_routing_keys="
            f"{[binding['routing_key'] for binding in self.producer_queue_bindings]}"
        )

    def _post(self, routing_key: str, message: dict) -> None:
        """
        Post message to the exchange configured on the producer instance
        :param routing_key: routing key to use on the published message.
            It must exist on the instance binding config
        :param message: body of the message to post
        :return: None
        """
        if not self.is_connected:
            self.connect()

        self.channel.basic_publish(
            exchange=self.publish_exchange,
            routing_key=routing_key,
            body=json.dumps(message),
            properties=self.publish_message_properties,
            mandatory=True,
        )

    # Retry posts which are not confirmed by the broker
    @retry((UnroutableError, NackError), tries=3, delay=1)
    def post(self, routing_key: str, message: dict, validate_routing_key: bool = True) -> None:
        """
        Retry as configured  message post to the exchange configured on the producer instance
        :param routing_key: routing key to use on the published message.
        It must exist on the instance binding config
        :param message: body of the message to post
        :param validate_routing_key: Boolean indicator to validate the routing key was configured during init
        :return: None
        """
        if validate_routing_key:
            self._validate_routing_key(routing_key)
        retry_call(self._post, fargs=[routing_key, message], **self.retry_config)

    def publish_message(self, message) -> None:
        """
        Publish a message using an object constructed with the talus.message.message_class decorator
        :return: None
        """
        if not hasattr(message, 'routing_key') and hasattr(message, 'dict'):
            raise TypeError("Argument 'message' must implement routing_key and dict methods.  "
                        "Recommend using talus.message.message_class decorator")
        self.post(message.routing_key(), message.dict())
